#include <string>
#include "UniSocket.hpp"
#include "SocketWrapperLinux.hpp"

#include <sys/socket.h>		// socket(), connect(), send(), recv()
#include <arpa/inet.h>		// inet_addr()
#include <unistd.h>			// close()
#include <netinet/in.h>		// sockaddr_in

#include <cstring>			// strerror
#include <cerrno>
#include <cstdio>			// sprintf()

// Client -----------------

using namespace std;

SocketWrapper::SocketWrapper(void) {}

SocketWrapper::SocketWrapper(
	const string & ip,
	const int & port,
	const string & end
) {
	_ip = ip;
	_end = end;
	_socket = ::socket(AF_INET, SOCK_STREAM , 0);
	if (_socket == -1)
		throw UniSocketException(strerror(errno));

	_address.sin_family = AF_INET;
	_address.sin_addr.s_addr = ::inet_addr(_ip.c_str());
	_address.sin_port = ::htons(port);
	
	int flag = ::connect(
		_socket,
		(struct sockaddr *) & _address,
		sizeof(_address)
	);

	if(flag == -1)
		throw UniSocketException(strerror(errno));
}

void SocketWrapper::send(const string & data) {
	string msg = data + _end;
	::write(_socket, msg.c_str(), msg.length());
}

string SocketWrapper::recv(void) {

	char block[256] = {};
	string result;
	int len;
	do {
		block[0] = '\0';
		len = ::read(_socket, block, 256);
		
		if(len != 0)
			result += block;
	} while(result.find(_end) == string::npos);
	
	// extract end - of - message
	result = result.substr(0, result.find(_end));
	return result;
}

void SocketWrapper::close(void) {
	::close(_socket);
}

// Server -----------------

SocketWrapper::SocketWrapper(
	const int & port,
	const int & queueSize,
	const string & end
) {
	_end = end;
	_socket = ::socket(AF_INET, SOCK_STREAM, 0);
	
	if(_socket == -1)
		throw UniSocketException(strerror(errno));

	int i = 1;
	::setsockopt(_socket, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i));

	_address.sin_family = AF_INET;
	_address.sin_addr.s_addr = INADDR_ANY;
	_address.sin_port = ::htons(port);
	
	int r = ::bind(
		_socket,
		(struct sockaddr *) & _address,
		sizeof(_address)
	);

	if(r == -1)
		throw UniSocketException(strerror(errno));

	r = ::listen(_socket, queueSize);
	
	if(r == -1)
		throw UniSocketException(strerror(errno));

}

SocketWrapper::SocketWrapper(
	const sockaddr_in & myAddress,
	const int & mySocket,
	const string & end
) {
	_socket = mySocket;
	_address = myAddress;
	_end = end;
	
	char buff[18];
	sprintf(buff,
		"%d.%d.%d.%d",
		int(myAddress.sin_addr.s_addr&0xFF),
		int((myAddress.sin_addr.s_addr&0xFF00)>>8),
		int((myAddress.sin_addr.s_addr&0xFF0000)>>16),
		int((myAddress.sin_addr.s_addr&0xFF000000)>>24)
	);
	_ip = buff;
}

SocketWrapper SocketWrapper::accept(void) {
	socklen_t addrlen = sizeof(struct sockaddr_in);
		
	int conn = ::accept(
		_socket,
		(struct sockaddr *) & _address,
		& addrlen
	);
	if(conn == -1)
		throw UniSocketException(strerror(errno));
	
	SocketWrapper so = SocketWrapper(_address, conn, _end);
	return so;
}
