#include <string>
#include "UniSocket.hpp"
#include "SocketWrapperWin.hpp"

#include <winsock2.h>
#include <ws2tcpip.h>		// inetPton()
#include <windows.h>

#include <cstring>			// strerror
#include <cerrno>
#include <cstdio>			// sprintf()

// Client -----------------

using namespace std;

SocketWrapper::SocketWrapper(void) {}

SocketWrapper::SocketWrapper(
	const string & ip,
	const int & port,
	const string & end
) {
	_ip = ip;
	_end = end;

	WORD wVersionRequested;
	WSADATA wsaData;

	wVersionRequested = MAKEWORD(2, 0); // Request WinSock v2.0
	if (::WSAStartup(wVersionRequested, &wsaData) != 0)
		throw UniSocketException("Unable to load WinSock DLL");

	_socket = (int) ::socket(AF_INET, SOCK_STREAM, 0);
	if (_socket == INVALID_SOCKET) {
		char err[300];
		strerror_s(err, 300, errno);
		throw UniSocketException(err);
	}

	memset(&_address, 0, sizeof(SOCKADDR_IN));
	_address.sin_family = AF_INET;
	_address.sin_addr.s_addr = inet_addr(_ip.c_str());
	_address.sin_port = htons(port);

	int flag = ::connect(
		_socket,
		(SOCKADDR *)& _address,
		sizeof(SOCKADDR)
	);

	if (flag == SOCKET_ERROR) {
		char err[300];
		strerror_s(err, 300, errno);
		throw UniSocketException(err);
	}
}

void SocketWrapper::send(const string & data) {
	string msg = data + _end;
	::send((SOCKET)_socket, msg.c_str(), msg.length(), 0);
}

string SocketWrapper::recv(void) {

	char block[256] = {};
	string result;
	int len;
	do {
		block[0] = '\0';
		len = ::recv(_socket, block, 256, 0);

		if (len != 0)
			result += block;
	} while (result.find(_end) == string::npos);

	// extract end - of - message
	result = result.substr(0, result.find(_end));
	return result;
}

void SocketWrapper::close(void) {
	closesocket((SOCKET)_socket);
}

// Server -----------------

SocketWrapper::SocketWrapper(
	const int & port,
	const int & queueSize,
	const string & end
) {
	_end = end;

	WORD wVersionRequested;
	WSADATA wsaData;

	wVersionRequested = MAKEWORD(2, 0); // Request WinSock v2.0
	if (::WSAStartup(wVersionRequested, &wsaData) != 0)
		throw UniSocketException("Unable to load WinSock DLL");

	_socket = (int) ::socket(AF_INET, SOCK_STREAM, 0);

	if (_socket == INVALID_SOCKET) {
		char err[300];
		strerror_s(err, 300, errno);
		throw UniSocketException(err);
	}

	int i = 1;
	::setsockopt(_socket, SOL_SOCKET, SO_REUSEADDR, (char *)&i, sizeof(i));

	memset(&_address, 0, sizeof(SOCKADDR_IN));
	_address.sin_family = AF_INET;
	_address.sin_addr.s_addr = ADDR_ANY;
	_address.sin_port = htons(port);

	int r = ::bind(
		_socket,
		(SOCKADDR *)& _address,
		sizeof(SOCKADDR_IN)
		);

	if (r == SOCKET_ERROR) {
		char err[300];
		strerror_s(err, 300, errno);
		throw UniSocketException(err);
	}

	r = ::listen(_socket, queueSize);

	if (r == SOCKET_ERROR) {
		char err[300];
		strerror_s(err, 300, errno);
		throw UniSocketException(err);
	}
}

SocketWrapper::SocketWrapper(
	const SOCKADDR_IN & myAddress,
	const int & mySocket,
	const string & end
) {
	_socket = mySocket;
	_address = myAddress;
	_end = end;
	
	char buff[18];
	sprintf(buff,
		"%d.%d.%d.%d",
		int(myAddress.sin_addr.s_addr&0xFF),
		int((myAddress.sin_addr.s_addr&0xFF00)>>8),
		int((myAddress.sin_addr.s_addr&0xFF0000)>>16),
		int((myAddress.sin_addr.s_addr&0xFF000000)>>24)
	);
	_ip = buff;
}

SocketWrapper SocketWrapper::accept(void) {
	int addrlen = sizeof(struct sockaddr_in);

	int conn = ::accept(
		_socket,
		(SOCKADDR *)& _address,
		&addrlen
	);
	if (conn == INVALID_SOCKET) {
		char err[300];
		strerror_s(err, 300, errno);
		throw UniSocketException(err);
	}

	SocketWrapper so = SocketWrapper(_address, conn, _end);
	return so;
}
